#ifndef _PW_PWTYPEDEF_H
#define _PW_PWTYPEDEF_H

/*
	ProWesS, PROGS Windowing System

	Definition of an external object type definition.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 7 January 1995
*/

#include "PWEvent.h"

#define PW_TYPEDEF_MODULE "ProWesS External Type Definition"

#define PW_TYPEDEF_FLAG 0x58505753	/* flag for PWXType */

#define PW_TYPE_EVENT_REGION	1	/* region associated with object type */
#define PW_TYPE_EVENT_KEYPRESS	2	/* keypress associated with object type */

typedef struct _PWTypeDef {
	int flag;						/* flag for external object 'XPWS' */

	PWType identifier;				/* type identifier, e.g. 'OUTL' */

	char event_type;				/* identifies type of event associated */
									/* with object of this type */

	/* the routines to handle calls which affect this type of object */
	Error (*PWCreate)(PWObject owner, PWObject *ret, PWEvent event, int *taglist);
	Error (*PWChange)(PWObject object, int *taglist);
	Error (*PWQuery)(PWObject object, int what, void *ret);
	Error (*PWRemove)(PWObject object);

	/* handlre routines */
	Error (*PWRegionEntry)(PWObject object);
	Error (*PWRegionExit)(PWObject object);
	Error (*PWHandle)(PWObject object, PWWait event);

	/* routine to define the parameters according to the configuration file */
	Error (*PWDefine)(char *name, char *value);

	/* draw routine for object (if it is drawable) */
	/* when an object redraws, it should automatically redraw all its children
	   which should be redrawn */
	Error (*PWDraw)(PWObject object);

	/* routines to handle the size aspects of the object */
	Error (*PWDetermineSize)(PWObject object);
	Error (*PWScale)(PWObject object, pt xinc, pt yinc);
};

#endif _PW_PWTYPEDEF_H
