/*
	ProWesS, PROGS Windowing System

	PWJobCreate 	create PWJob data for the current job (if not yet exists)

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 January 1995
*/

#include "ProWesS_h"
#include "PWJob_h"

#include "job_h"
#include "atomic_h"

/* global list with job information, double linked list with sentinel */
/* this list should only be accessed from atomic routines !! */
static PWJob ProWesS_jobs_sentinel =
			{ &ProWesS_jobs_sentinel, &ProWesS_jobs_sentinel, -1, NULL};
static PWJob *ProWesS_jobs=&ProWesS_jobs_sentinel;


static Error PWJobSearch(Job id, PWJob **jobinfo)
{
	PWJob *curr=ProWesS_jobs->next;

	while (curr->job!=-1 && curr->job!=id) curr=curr->next;
	if (curr->job==id)
	{
		*jobinfo=curr;
		return ERR_OK;
	}
	else
	{
		return ERR_ITNF;
	}
}

static void PWJobUnlink(void *base)
{
	PWJob *jobinfo=(PWJob *)base;
	jobinfo->next->prev=jobinfo->prev;
	jobinfo->prev->next=jobinfo->next;
}

static Error PWJobLink(PWJob *jobinfo)
{
	PWJob *curr=ProWesS_jobs;

	jobinfo->next=curr->next;
	jobinfo->prev=curr;
	curr->next->prev=jobinfo;
	curr->next=jobinfo;
	MEMWhenZapped(jobinfo,&PWJobUnlink,NULL);

	return ERR_OK;
}

Error PWJobCreate(PWJob **jobinfo)
{
	Error err; Job me;

	/* see if there already was a PWJob for this job */
	JOBMe(&me);
	err=ATOMICCall(PWJobSearch,2,me,jobinfo);
	if (!err) return ERR_OK;

	/* allocate PWJob */
	err=MEMAllocate(sizeof(PWJob),jobinfo);
	if (err) return err;

	/* initialise PWJob (not too much work) */
	(*jobinfo)->job=me;
	/* the other entries are NULL, as MEMAllocate clears the memory ... */

	/* link PWJob in list of applications which use ProWesS */
	return ATOMICCall(PWJobLink,1,*jobinfo);
}
